/* setgamma.c  Stuart Levy, Geometry Center, University of Minnesota  1995 */
/* cc  -o setgamma  setgamma.c -lgl -lm */

#include <gl/gl.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>

char Usage[] = "\
Usage: %s [-b brighten] [-B prebrighten] [gamma]\n\
Uses the SGI hardware gamma mapping to adjust the brightness and/or gamma\n\
of the display.  With no arguments, resets brightness and gamma to their\n\
defaults.\n\
  ``brightness'' multiplies all displayed screen values by the given factor;\n\
	may be a single number or 3 comma-separated ones for independent R,G,B\n\
	brightening.  Default is ``-b 1'' a.k.a. ``-b 1,1,1''.\n\
  ``gamma'' sets the exponent of a power-law mapping of pixel values to	\n\
	screen values, and may also be either one or three comma-separated\n\
	numbers.  Larger values brighten faint things, lightening\n\
	shadows, and making colors less saturated.  The default is the\n\
	systemwide value as set by gamma(1), normally ``-g 1.7''.\n\
	Typically -g 2.4 gives linear relation from pixel value to brightness.\n\
\n\
Unlike /usr/sbin/gamma, this program\n\
  - does not require special privileges (while only root can run ``gamma'')\n\
  - makes only a non-permanent setting, which reverts to the system default\n\
    typically at logout.\n\
";

void screengamma(float gamma[3], float max[3], float prescale[3], float rewhite[3])
{
    static float one[3] = { 1,1,1 };
    short ramp[3][256];
    float framp[256];
    int i, c;
    static int fullscale = 255;

    if(max == NULL) max = one;
    printf(gamma[1] == gamma[0] && gamma[2] == gamma[0]
	? "Setting gamma to %g" : "Setting gamma: red %g green %g blue %g",
	gamma[0], gamma[1], gamma[2]);
    if(max[0] != 1 || max[1] != 1 || max[2] != 1) {
	printf(max[0] == max[1] && max[0] == max[2]
	 ? ", max RGB %g" : ", max R %g G %g B %g",
	 max[0], max[1], max[2]);
    }
    printf("\n");

    for(c = 0; c < 3; c++) {
	if(c>0 && gamma[c]==gamma[c-1] && max[c] == max[c-1] && prescale[c] == prescale[c-1]) {
	    memcpy(&ramp[c][0], &ramp[c-1][0], 256*sizeof(short));
	} else {
	    for(i = 0; i < 256; i++)
		framp[i] = pow(prescale[c]*i/255., 1./gamma[c]);

	    fullscale = 255 * max[c];
	    for(i = 0; i < 256; i++) {
		float v = fullscale<0 ? (255 + fullscale*(framp[i]))
				      : fullscale * framp[i];
		ramp[c][i] = (v<0) ? 0 : (v>255) ? 255 : v;
	    }
	}
    }
    if(rewhite) {
	ramp[0][255] = 255*rewhite[0];
	ramp[1][255] = 255*rewhite[1];
	ramp[2][255] = 255*rewhite[2];
    }
    gammaramp(ramp[0], ramp[1], ramp[2]);
}

main(int argc, char *argv[]) {
    float gamma[3];
    float max[3];
    float prescale[3];
    float white[3];
    int gotgamma = 0;
    int setwhite = 0;
    int k;
    char *prog = argv[0];

    gamma[0] = gamma[1] = gamma[2] = 1.7;
    max[0] = max[1] = max[2] = 1;
    prescale[0] = prescale[1] = prescale[2] = 1;

    while(argc > 2 && argv[1][0] == '-') {
        if(argv[1][1] == 'b') {
	    k = sscanf(argv[2], "%f%*c%f%*c%f", &max[0], &max[1], &max[2]);
	    if(k < 2) max[1] = max[0];
	    if(k < 3) max[2] = max[1];
	    argc -= 2, argv += 2;

	} else if(argv[1][1] == 'B') {
	    k = sscanf(argv[2], "%f%*c%f%*c%f",
		&prescale[0], &prescale[1], &prescale[2]);
	    if(k < 2) prescale[1] = prescale[0];
	    if(k < 3) prescale[2] = prescale[1];
	    argc -= 2, argv += 2;

	} else if(argv[1][1] == 'w') {
	    k = sscanf(argv[2], "%f%*c%f%*c%f", &white[0], &white[1], &white[2]);
	    if(k < 2) white[1] = white[0];
	    if(k < 3) white[2] = white[1];
	    setwhite = 1;
	    argc -= 2, argv += 2;

	} else if(argv[1][1] == 'g') {
	    switch( sscanf(argv[2], "%f%*c%f%*c%f", &gamma[0], &gamma[1], &gamma[2]) ) {
	    case 0: case -1:
		fprintf(stderr, Usage, prog);
		exit(1);
	    case 1:  gamma[1] = gamma[0];
	    case 2:  gamma[2] = gamma[1];
	    }
	    gotgamma = 1;
	    argc -= 2, argv += 2;
	} else {
	    break;
	}
    }

    if(argc > 1) {
	switch( sscanf(argv[1], "%f%*c%f%*c%f", &gamma[0], &gamma[1], &gamma[2]) ) {
	case 0: case -1:
	    fprintf(stderr, Usage, prog);
	    exit(1);
	case 1:  gamma[1] = gamma[0];	/* ... and fall into ... */
	case 2:  gamma[2] = gamma[1];
	}
	if(argc > 2) gamma[1] = gamma[2] = atof(argv[2]);
	if(argc > 3) gamma[2] = atof(argv[3]);
	gotgamma = 1;
    }
    if(!gotgamma) {
	FILE *gamfile = fopen("/etc/config/system.glGammaVal", "r");
	if(gamfile)
	    fscanf(gamfile, "%f", &gamma[0]);
	gamma[1] = gamma[2] = gamma[0];
    }
	
    foreground();
    noport();
    winopen("");
    screengamma( gamma, max, prescale, setwhite ? white : NULL );
    exit(0);
}

